@@ -33,8 +33,6 @@ import tempfile
 import shutil
 import urllib.parse
 
-from debian import debian_support
-
 import adtlog
 import VirtSubproc
 
@@ -48,7 +46,7 @@ class Testbed:
                  setup_commands=[], setup_commands_boot=[], add_apt_pockets=[],
                  copy_files=[], pin_packages=[], add_apt_sources=[],
                  add_apt_releases=[], apt_default_release=None,
-                 enable_apt_fallback=True):
+                 enable_apt_fallback=True, shell_fail=False, needs_internet='run'):
         self.sp = None
         self.lastsend = None
         self.scratch = None
@@ -79,6 +77,8 @@ class Testbed:
         self.eatmydata_prefix = []
         self.apt_pin_for_releases = []
         self.enable_apt_fallback = enable_apt_fallback
+        self.needs_internet = needs_internet
+        self.shell_fail = shell_fail
         self.nproc = None
         self.cpu_model = None
         self.cpu_flags = None
@@ -95,6 +95,9 @@ class Testbed:
         return self.user or 'root'
 
     def start(self):
+        # log date at least once; to ease finding it
+        adtlog.info('starting date: %s' % time.strftime('%Y-%m-%d'))
+
         # are we running from a checkout?
         root_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
         if os.path.exists(os.path.join(root_dir, '.git')):
@@ -135,6 +138,7 @@ class Testbed:
             self.sp.stdin.close()
             ec = self.sp.wait()
         if ec:
+            self.command('auxverb_debug_fail')
             self.bomb('testbed gave exit status %d after quit' % ec)
         self.sp = None
 
@@ -209,6 +213,8 @@ class Testbed:
             if rc:
                 # setup scripts should exit with 100 if it's the package's
                 # fault, otherwise it's considered a transient testbed failure
+                if self.shell_fail:
+                    self.run_shell()
                 if rc == 100:
                     self.badpkg('testbed boot setup commands failed with status 100')
                 else:
@@ -221,6 +227,8 @@ class Testbed:
         self.recommends_installed = False
         self.exec_cmd = list(map(urllib.parse.unquote, self.command('print-execute-command', (), 1)[0].split(',')))
         self.caps = self.command('capabilities', (), None)
+        if self.needs_internet in ['try', 'run']:
+            self.caps.append('has_internet')
         adtlog.debug('testbed capabilities: %s' % self.caps)
         for c in self.caps:
             if c.startswith('downtmp-host='):
@@ -331,7 +339,7 @@ class Testbed:
                                                   [self._get_default_release() + '-updates'])
         if self.add_apt_releases:
             get_mirror_and_components = '''
-            sed -rn 's/^(deb|deb-src) +(\\[.*\\] *)?((http|https|file):[^ ]*) +([^ -]+) +(.*)$/\\2\\3 \\6/p' \\
+            sed -rn 's/^(deb|deb-src) +(\\[.*\\] *)?((http|https|file):[^ ]*) +([^ ]+) +(.*)$/\\2\\3 \\6/p' \\
                 /etc/apt/sources.list `ls /etc/apt/sources.list.d/*.list 2>/dev/null|| true` | head -1
             '''
             mirror_and_components = self.check_exec(['sh', '-ec', get_mirror_and_components], stdout=True).split()
@@ -365,6 +373,8 @@ class Testbed:
             if rc:
                 # setup scripts should exit with 100 if it's the package's
                 # fault, otherwise it's considered a transient testbed failure
+                if self.shell_fail:
+                    self.run_shell()
                 if rc == 100:
                     self.badpkg('testbed setup commands failed with status 100')
                 else:
@@ -422,7 +432,7 @@ class Testbed:
         adtlog.debug('%s %s' % (_type.__name__, m))
         raise _type(m)
 
-    def send(self, string):
+    def send(self, string, debug_if_fails=True):
         try:
             adtlog.debug('sending command to testbed: ' + string)
             self.sp.stdin.write(string)
@@ -430,11 +440,16 @@ class Testbed:
             self.sp.stdin.flush()
             self.lastsend = string
         except Exception as e:
+            # command() calls back into send() - avoid potential infinite
+            # recursion
+            if debug_if_fails:
+                self.command('auxverb_debug_fail', debug_if_fails=False)
             self.bomb('cannot send to testbed: %s' % e)
 
     def expect(self, keyword, nresults):
         line = self.sp.stdout.readline()
         if not line:
+            self.command('auxverb_debug_fail')
             self.bomb('unexpected eof from the testbed')
         if not line.endswith('\n'):
             self.bomb('unterminated line from the testbed')
@@ -445,9 +460,11 @@ class Testbed:
             self.bomb('unexpected whitespace-only line from the testbed')
         if ll[0] != keyword:
             if self.lastsend is None:
+                self.command('auxverb_debug_fail')
                 self.bomb("got banner `%s', expected `%s...'" %
                           (line, keyword))
             else:
+                self.command('auxverb_debug_fail')
                 self.bomb("sent `%s', got `%s', expected `%s...'" %
                           (self.lastsend, line, keyword))
         ll = ll[1:]
@@ -457,7 +474,7 @@ class Testbed:
                       (self.lastsend, line, len(ll), nresults))
         return ll
 
-    def command(self, cmd, args=(), nresults=0, unquote=True):
+    def command(self, cmd, args=(), nresults=0, unquote=True, debug_if_fails=True):
         # pass args=[None,...] or =(None,...) to avoid more url quoting
         if type(cmd) is str:
             cmd = [cmd]
@@ -466,7 +483,7 @@ class Testbed:
         else:
             args = list(map(urllib.parse.quote, args))
         al = cmd + args
-        self.send(' '.join(al))
+        self.send(' '.join(al), debug_if_fails=debug_if_fails)
         ll = self.expect('ok', nresults)
         if unquote:
             ll = list(map(urllib.parse.unquote, ll))
@@ -520,6 +537,7 @@ class Testbed:
                 adtlog.error(msg)
                 raise
             else:
+                self.command('auxverb_debug_fail')
                 self.bomb(msg)
 
         adtlog.debug('testbed command exited with code %i' % proc.returncode)
@@ -756,176 +774,6 @@ Description: satisfy autopkgtest test dependencies
 
         self.execute(['dpkg', '--purge', 'autopkgtest-satdep'])
 
-    def install_tmp(self, deps, recommends=False):
-        '''Unpack dependencies into temporary directory
-
-        This is a fallback if the testbed does not have root privileges or a
-        writable file system, and will only work for packages that can be
-        used from a different directory with PATH, LD_LIBRARY_PATH, PYTHONPATH
-        etc. set.
-
-        Sets/updates self.install_tmp_env to necessary variables.
-        '''
-        unsupported = []
-        pkg_constraints = {}  # pkg -> (relation, version)
-
-        # parse deps into pkg_constraints
-        dep_re = re.compile(
-            r'(?P<p>[a-z0-9+-.]+)\s*'
-            r'(\((?P<r><<|<=|>=|=|>>)\s*(?P<v>[^\)]*)\))?$')
-        for dep in deps.split(','):
-            dep = dep.strip()
-            if not dep:
-                continue  # trailing comma
-            m = dep_re.match(dep)
-            if not m:
-                unsupported.append(dep)
-                continue
-            pkg_constraints[m.group('p')] = (m.group('r'), m.group('v'))
-
-        adtlog.debug('install_tmp: "%s" -> %s, unsupported: %s' %
-                     (deps, pkg_constraints, unsupported))
-
-        if unsupported:
-            adtlog.warning('The following dependencies cannot be handled in '
-                           'reduced "unpack to temporary directory" mode: ' +
-                           ', '.join(unsupported))
-
-        # simulate installation, grab packages, and check constraints
-        (rc, out, _) = self.execute(['apt-get', '--quiet', '--simulate', '--no-remove',
-                                     '-o', 'Debug::pkgProblemResolver=true',
-                                     '-o', 'Debug::NoLocking=true',
-                                     '-o', 'APT::Install-Recommends=%s' % recommends,
-                                     '-o', 'APT::Get::Show-User-Simulation-Note=False',
-                                     'install'] + list(pkg_constraints),
-                                    stdout=subprocess.PIPE)
-        if rc != 0:
-            self.badpkg('Test dependencies are unsatisfiable. A common reason is '
-                        'that your testbed is out of date with respect to the '
-                        'archive, and you need to use a current testbed, or '
-                        'try "--setup-commands ro-apt-update".')
-
-        def check_constraint(pkg, ver):
-            constraint = pkg_constraints.get(pkg, (None, None))
-            if constraint[0] is None:
-                return True
-            comp = debian_support.version_compare(ver, constraint[1])
-            if constraint[0] == '<<':
-                return comp < 0
-            if constraint[0] == '<=':
-                return comp <= 0
-            if constraint[0] == '==':
-                return comp == 0
-            if constraint[0] == '>=':
-                return comp >= 0
-            if constraint[0] == '>>':
-                return comp > 0
-            raise ValueError('invalid dependency version relation %s' % constraint[0])
-
-        to_install = []
-        for line in out.splitlines():
-            if not line.startswith('Inst '):
-                continue
-            fields = line.split()
-            pkg = fields[1]
-            if fields[2].startswith('('):
-                ver = fields[2][1:]
-            elif fields[3].startswith('('):
-                ver = fields[3][1:]
-            else:
-                raise ValueError('Cannot parse line: %s' % line)
-            # ignore Python 2 stuff, with PYTHONPATH we can only support one
-            # Python major version (3)
-            if pkg.startswith('python-') or pkg.startswith('libpython-') or \
-               'python2.' in pkg or pkg == 'python':
-                adtlog.warning('Ignoring Python 2.x dependency %s, not '
-                               'supported in unpack only mode' % pkg)
-                continue
-            if not check_constraint(pkg, ver):
-                self.badpkg('test dependency %s (%s %s) is unsatisfiable: available version %s' %
-                            (pkg, pkg_constraints[pkg][0], pkg_constraints[pkg][1], ver))
-            to_install.append(pkg)
-
-        adtlog.debug('install_tmp: packages to install: %s' % ' '.join(to_install))
-
-        if not to_install:
-            # we already have everything, all good
-            return
-
-        adtlog.warning('virtualisation system does not offer root or writable '
-                       'testbed; unpacking dependencies to temporary dir, '
-                       'which will only work for some packages')
-
-        # download and unpack all debs
-        script = r'''d=%(t)s/deps
-mkdir -p $d; cd $d
-apt-get download %(pkgs)s >&2
-for p in *.deb; do dpkg-deb --extract $p .; rm $p; done
-
-# executables
-echo PATH=$d/sbin:$d/bin:$d/usr/sbin:$d/usr/bin:$d/usr/games:$PATH
-
-# shared libraries / Qt plugins
-l=""
-q=""
-for candidate in $(find $d -type d \( -name 'lib' -o -path '*/lib/*-linux-*' \)); do
-    [ -z "$(ls $candidate/*.so $candidate/*.so.* 2>/dev/null)" ] || l="$candidate:$l"
-    [ -z "$(ls $candidate/lib*qt*.so* 2>/dev/null)" ] || q="$candidate:$q"
-done
-[ -z "$l" ] || echo LD_LIBRARY_PATH=$l${LD_LIBRARY_PATH:-}
-[ -z "$q" ] || echo QT_PLUGIN_PATH="$q"
-
-# ImageMagick needs some hacks to make python[3]-wand find its library
-l=""
-for ml in $(ls usr/lib/*-linux-*/libMagick*.so.* 2>/dev/null); do
-    if [ -L $ml ]; then continue; fi
-    l=$(dirname $ml)
-    ln -sf $(basename "$ml") "${ml%%.so.*}.so"
-done
-if [ -n "$l" ]; then
-    [ -d "$l/lib" ] || ln -sf . "$l/lib"
-    echo MAGICK_HOME="$d/$l"
-fi
-
-# Python modules
-p=""
-for candidate in $d/usr/lib/python3*/dist-packages; do
-    [ ! -d $candidate ] || p="$candidate:$p"
-done
-[ -z "$p" ] || echo PYTHONPATH=$p${PYTHONPATH:-}
-
-# Perl modules
-p=""
-for candidate in $d/usr/share/perl* $d/usr/lib/perl5 $d/usr/lib/*/perl5/*; do
-    [ ! -d $candidate ] || p="$candidate:$p"
-done
-[ -z "$p" ] || echo PERL5LIB=$p${PERL5LIB:-}
-
-# gobject-introspection
-l=""
-if [ -d $d/usr/lib/girepository-1.0 ]; then
-    l=$d/usr/lib/girepository-1.0
-fi
-for candidate in $(find $d -type d -path '*/usr/lib/*/girepository-*'); do
-    [ -z "$(ls $candidate/*.typelib 2>/dev/null)" ] || l="$candidate:$l"
-done
-[ -z "$l" ] || echo GI_TYPELIB_PATH="$l:${GI_TYPELIB_PATH:-}"
-
-# udev rules
-if [ -n "$(ls $d/lib/udev/rules.d/*.rules 2>/dev/null)" ] && [ -w /run/udev ]; then
-    mkdir -p /run/udev/rules.d
-    cp $d/lib/udev/rules.d/*.rules /run/udev/rules.d/
-    udevadm control --reload
-    udevadm trigger || true
-fi
-''' % {'t': self.scratch, 'pkgs': ' '.join(to_install)}
-        (rc, out, _) = self.execute(['sh', '-euc', script],
-                                    stdout=subprocess.PIPE, kind='install')
-        if rc != 0:
-            self.bomb('failed to download and unpack test dependencies')
-        self.install_tmp_env = [l.strip() for l in out.splitlines() if l]
-        adtlog.debug('install_tmp: env is now %s' % self.install_tmp_env)
-
     def install_click(self, clickpath):
         # copy click into testbed
         tp = Path(self, clickpath, os.path.join(
@@ -1066,7 +914,22 @@ fi
         if can_apt_get:
             self.install_apt(deps, recommends, shell_on_failure, synth_deps)
         else:
-            self.install_tmp(deps, recommends)
+            has_dpkg_checkbuilddeps = self.execute(
+                ['which', 'dpkg-checkbuilddeps'],
+                stdout=subprocess.DEVNULL,
+                stderr=subprocess.DEVNULL
+            )[0] == 0
+            if has_dpkg_checkbuilddeps:
+                rc, _, err = self.execute(
+                    ['dpkg-checkbuilddeps', '-d', deps, "/dev/null"],
+                    stdout=subprocess.PIPE,
+                    stderr=subprocess.PIPE
+                )
+                if rc != 0:
+                    missing = re.sub('dpkg-checkbuilddeps: error: Unmet build dependencies: ', '', err)
+                    self.badpkg('test dependencies missing: %s' % missing)
+            else:
+                adtlog.warning('test dependencies (%s) are not fully satisfied, but continuing anyway since dpkg-checkbuilddeps it not available to determine which ones are missing.' % deps)
 
     def run_shell(self, cwd=None, extra_env=[]):
         '''Run shell in testbed for debugging tests'''
@@ -1270,7 +1133,10 @@ fi
         elif rc == 77 and 'skippable' in test.restrictions:
             test.set_skipped('exit status 77 and marked as skippable')
         elif rc != 0:
-            test.failed('non-zero exit status %d' % rc)
+            if 'needs-internet' in test.restrictions and self.needs_internet == 'try':
+                test.set_skipped("Failed, but test has needs-internet and that's not guaranteed")
+            else:
+                test.failed('non-zero exit status %d' % rc)
         elif se_size != 0 and 'allow-stderr' not in test.restrictions:
             with open(se.host, encoding='UTF-8', errors='replace') as f:
                 stderr_top = f.readline().rstrip('\n \t\r')
@@ -1374,7 +1240,7 @@ fi
         if self.default_release is None:
 
             script = 'SRCS=$(ls /etc/apt/sources.list /etc/apt/sources.list.d/*.list 2>/dev/null|| true); '
-            script += r'''sed -rn '/^(deb|deb-src) +(\[.*\] *)?(http|https|file):/ { s/\[.*\] +//; s/^[^ ]+ +[^ ]* +([^ -]+) +.*$/\1/p }' $SRCS | head -n1'''
+            script += r'''sed -rn '/^(deb|deb-src) +(\[.*\] *)?(http|https|file):/ { s/\[.*\] +//; s/^[^ ]+ +[^ ]* +([^ ]+) +.*$/\1/p }' $SRCS | head -n1'''
             self.default_release = self.check_exec(['sh', '-ec', script], stdout=True).strip()
 
         return self.default_release
@@ -1484,6 +1350,10 @@ class TempPath(Path):
 
     These are only guaranteed to exit within one testbed run.
     '''
+
+    # private; used to make sure the names of autocleaned files don't collide
+    _filename_prefix = 1
+
     def __init__(self, testbed, name, is_dir=False, autoclean=True):
         '''Create a temporary Path object.
 
@@ -1507,6 +1377,10 @@ class TempPath(Path):
         else:
             host = testbed.output_dir
         self.autoclean = autoclean
+        if autoclean:
+            name = str(self._filename_prefix) + '-' + name
+            TempPath._filename_prefix += 1
+
         Path.__init__(self, testbed, os.path.join(host, name),
                       os.path.join(testbed.scratch, name),
                       is_dir)
