@@ -44,7 +44,7 @@ known_restrictions = ['rw-build-tree', 'breaks-testbed', 'needs-root',
                       'build-needed', 'allow-stderr', 'isolation-container',
                       'isolation-machine', 'needs-recommends', 'needs-reboot',
                       'flaky', 'skippable', 'superficial',
-                      'skip-not-installable']
+                      'skip-not-installable', 'needs-internet']
 
 
 class Unsupported(Exception):
@@ -187,6 +187,12 @@ class Test:
                               'Test needs to reboot testbed but testbed does '
                               'not provide reboot capability')
 
+        if 'needs-internet' in self.restrictions and \
+           'has_internet' not in caps:
+            raise Unsupported(self.name,
+                              'Test needs unrestricted internet access but testbed does '
+                              'not provide it')
+
 #
 # Parsing for Debian source packages
 #
@@ -296,9 +302,9 @@ def _debian_build_deps_from_source(srcdir, testbed_arch):
 
 
 dep_re = re.compile(
-    r'(?P<package>[a-z0-9+-.]+)(?::native)?\s*'
+    r'(?P<package>[a-z0-9+-.]+)(?::[a-z0-9_-]+)?\s*'
     r'(\((?P<relation><<|<=|>=|=|>>)\s*(?P<version>[^\)]*)\))?'
-    r'(\s*\[(?P<arch>[[a-z0-9+-.! ]+)\])?$')
+    r'(\s*\[(?P<arch>[a-z0-9+-.! ]+)\])?$')
 
 
 def _debian_check_dep(testname, dep):
@@ -393,11 +399,11 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch):
                 if s:
                     synthdep_alternatives.append(s)
             if synthdep_alternatives:
-                    adtlog.debug('marked alternatives %s as a synthesised dependency' % synthdep_alternatives)
-                    if len(synthdep_alternatives) > 1:
-                        synthdeps.append(synthdep_alternatives)
-                    else:
-                        synthdeps.append(synthdep_alternatives[0])
+                adtlog.debug('marked alternatives %s as a synthesised dependency' % synthdep_alternatives)
+                if len(synthdep_alternatives) > 1:
+                    synthdeps.append(synthdep_alternatives)
+                else:
+                    synthdeps.append(synthdep_alternatives[0])
             deps.append(alt_group_str)
 
     return (deps, synthdeps)
@@ -447,16 +453,30 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
     tests = []
     if not control_path:
         control_path = os.path.join(srcdir, 'debian', 'tests', 'control')
+        dtc_exists = os.path.exists(control_path)
+        try_autodep8 = False
 
-        if not os.path.exists(control_path):
-            if auto_control:
-                control = _autodep8(srcdir)
-                if control is None:
-                    return ([], False)
-                control_path = control.name
+        if auto_control:
+            if not dtc_exists:
+                try_autodep8 = True
             else:
-                adtlog.debug('auto_control is disabled, no tests')
+                dcontrol_path = os.path.join(srcdir, 'debian', 'control')
+                for record in parse_rfc822(dcontrol_path):
+                    testsuite = record.get('Testsuite', '')
+                    if 'autopkgtest-pkg-' in testsuite:
+                        try_autodep8 = True
+                    # We only want to look at the source section
+                    break
+
+        if try_autodep8:
+            control = _autodep8(srcdir)
+            if control is not None:
+                control_path = control.name
+            elif not dtc_exists:
                 return ([], False)
+        elif not dtc_exists:
+            adtlog.debug('auto_control is disabled, and no regular tests')
+            return ([], False)
 
     for record in parse_rfc822(control_path):
         command = None
@@ -498,12 +518,17 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                         '*', 'test-name feature incompatible with Tests')
                 test_dir = record.get('Tests-directory', 'debian/tests')
 
-                _debian_check_unknown_fields(test_names[0], record)
                 for n in test_names:
-                    test = Test(n, os.path.join(test_dir, n), None,
-                                restrictions, features, depends, [], [], synth_depends)
-                    test.check_testbed_compat(testbed_caps)
-                    tests.append(test)
+                    try:
+                        _debian_check_unknown_fields(n, record)
+                        test = Test(n, os.path.join(test_dir, n), None,
+                                    restrictions, features, depends, [], [], synth_depends)
+                        test.check_testbed_compat(testbed_caps)
+                    except Unsupported as u:
+                        u.report()
+                        some_skipped = True
+                    else:
+                        tests.append(test)
             elif 'Test-command' in record:
                 command = record['Test-command']
                 (depends, synth_depends) = _parse_debian_depends(
